/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * blit.h
 *
 *  Created on: 08/09/2010
 *      Author: raster
 */

#ifndef TTOS_H_
#define TTOS_H_

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>

#include "sharemem.h"

typedef struct _TTOS_Surface {
	int32_t w, h;
	uint16_t *pixels;
	uint8_t *alpha;
	bool uses_shared_mem;
	void *private;
} TTOS_Surface;

TTOS_Surface *blit_create_surface(uint32_t w, uint32_t h, bool alpha);
TTOS_Surface *blit_create_surface_from_shared(struct _shared_mem *sharemem, bool alpha);
void blit_destroy_surface(TTOS_Surface *surface);
int32_t blit_blend(TTOS_Surface *origin, int32_t ox, int32_t oy, int32_t ow, int32_t oh, TTOS_Surface *destination, int32_t dx, int32_t dy, int32_t crop_x, int32_t crop_y, int32_t crop_w, int32_t crop_h, bool do_copy);
int32_t blit_fill(TTOS_Surface *destination, int32_t dx, int32_t dy, uint32_t ow, uint32_t oh, struct ttos_color *color, bool keep_bg);
int32_t blit_multipixel(TTOS_Surface *destination, uint32_t x, uint32_t y, struct ttos_color *color);
int32_t blit_pixel(TTOS_Surface *destination, uint32_t x, uint32_t y, struct ttos_color *color);
int32_t blit_erase_surface(TTOS_Surface *destination);

#endif /* TTOS_H_ */
