/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * main.c
 *
 *  Created on: 05/09/2010
 *      Author: raster
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <poll.h>
#include <unistd.h>
#include <string.h>

#include "minibus/minibus.h"
#include "portable.h"
#include "blit.h"
#include "global_vars.h"
#include "clients.h"
#include "bitmaps.h"
#include "tabs.h"
#include "widgets.h"
#include "other.h"
#include "timer.h"
#include "fonts.h"
#include "test.h"
#include "uid.h"
#include "sharemem.h"
#include "ipc.h"

extern char **environ;

uint32_t cb_new_connection(struct minibus_connection *connection,void *data) {

	struct _client_data *object;
	uint32_t client_uid;

	object=clients_new(connection,NULL);
	if (object==NULL) {
		return -1;
	}

	client_uid=minibus_get_uid(connection);

	minibus_list_add(&clients,object);

	// if there's no window manager, set this program as the current visible program
	if (!window_manager) {
		clients_do_move_to_fg(client_uid);
	}

	return 0;
}

uint32_t cb_closed_connection(struct minibus_connection *connection,void *data) {

	struct _client_data *client;
	struct minibus_connection *wm_conn;
	uint32_t uid;
	uint8_t window;

	client=minibus_get_connection_data(connection);

	window=client->window;
	uid=minibus_get_uid(connection);

	if (client->window_manager) {
		window_manager=NULL;
		minibus_exit_loop(main_queue);
	}

	if (-1==minibus_list_remove_data(&clients,client,true)) {
		return -1;
	}

	if (windows[window].fg_client==client) {
		windows[window].fg_client=NULL;
	}

	/*
	 * Send a "destroyed client" message to the window manager
	 */
	if (window_manager) {
		wm_conn=window_manager->conn;
		minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_DESTROYED_CLIENT);
		minibus_write_long(write_conn,uid);
		if (windows[window].fg_client) {
			minibus_write_byte(write_conn,0); // there's still a foreground client in this window
		} else {
			minibus_write_byte(write_conn,1); // now this window has not a foreground client
		}
		minibus_write_byte(write_conn,window);
		minibus_server_send(write_conn,wm_conn,false);
	}

	minibus_set_connection_data(connection,NULL);

	return 0;
}

int32_t main_refresh(struct minibus_connection *connection) {

	uint32_t loop;
	struct _window *wpointer;
	struct _client_data *client;

	if (connection) {
		client=minibus_get_connection_data(connection);

		if(windows[client->window].fg_client->conn!=connection) {
			return 0;
		}
	}

	blit_erase_surface(base);

	for(loop=0,wpointer=windows;loop<n_windows;loop++,wpointer++) {
		if ((!wpointer)||(!wpointer->fg_client)||(!wpointer->visible)) {
			continue;
		}
		tabs_paint_tab(wpointer->fg_client->show_tab);
	}
	ttos_dump_screen(base);
	return 0;
}


uint32_t main_refresh_cb(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	main_refresh(connection);
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;

}


bool main_adjust_coordinates(struct ttos_tssample *sample1,struct ttos_tssample *sample2, uint32_t cwindow, struct _widget *widget) {

	struct _tab *ctab;
	uint32_t wx,wy;

	ctab=widget->parent_tab;
	wx=windows[cwindow].ox;
	wy=windows[cwindow].oy;

	sample2->x=sample1->x-wx;
	sample2->y=sample1->y-wy;
	sample2->ox=sample1->ox-wx;
	sample2->oy=sample1->oy-wy;
	if (!widget->h_fixed) {
		sample2->x-=ctab->x_offset;
		sample2->ox-=ctab->x_offset;
	}
	if (!widget->v_fixed) {
		sample2->y-=ctab->y_offset;
		sample2->oy-=ctab->y_offset;
	}
	sample2->x-=widget->ox;
	sample2->ox-=widget->ox;
	sample2->y-=widget->oy;
	sample2->oy-=widget->oy;
	sample2->type=sample1->type;
	return true;
}


void touchscreen_cb (int fd, short revents,void *data) {

	struct ttos_tssample sample,sample2;
	struct minibus_list_element *element;
	struct _widget *tmp_widget;
	struct _tab *ctab;
	uint32_t loop,loop2;
	int32_t wx,wy,tmpx,tmpy;
	static bool double_click=false;
	static struct timeval lasttime;
	static struct _widget *last_widget=NULL;

	ttos_read_ts(ts,&sample);

	if (sample.type==TTOS_TOUCH_START_CLICK) {
		lasttime.tv_sec=sample.tv.tv_sec;
		lasttime.tv_usec=sample.tv.tv_usec;

		// remove any timer and add a new one
		timer_check_event(TTOS_TIMER_LONG_CLICK,true,true,750);

		touch_window=0xFFFFFFFF;
		for(loop2=0;loop2<n_windows;loop2++) {
			loop=n_windows-loop2-1; // From top to bottom
			if (!windows[loop].visible) {
				continue;
			}
			if ((sample.x>=windows[loop].ox)&&(sample.x<=windows[loop].mx)&&
			(sample.y>=windows[loop].oy)&&(sample.y<=windows[loop].my)) {
				touch_window=loop;
				touch_widget=NULL;
				if (!windows[loop].fg_client)
					continue;
				if (!windows[loop].fg_client->show_tab)
					continue;
				if (!windows[loop].fg_client->show_tab->sensitive)
					continue;
				ctab=windows[loop].fg_client->show_tab;

				ctab->xtmp=ctab->x_offset;
				ctab->ytmp=ctab->y_offset;
				ctab->valid_tmp=true;
				wx=windows[loop].ox;
				wy=windows[loop].oy;

				for(element=ctab->widgets.last;element!=NULL;element=element->prev) {
					tmp_widget=(struct _widget *)element->data;
					if (!tmp_widget->visible) {//||(!tmp_widget->event_cb)) {
						continue;
					}
					tmpx=tmp_widget->ox+wx;
					tmpy=tmp_widget->oy+wy;
					if (!tmp_widget->h_fixed)
						tmpx+=ctab->x_offset;
					if (!tmp_widget->v_fixed)
						tmpy+=ctab->y_offset;
					if ((sample.x>=tmpx)&&(sample.x<(tmpx+tmp_widget->w))&&(sample.y>=tmpy)&&(sample.y<(tmpy+tmp_widget->h))) {
						touch_widget=tmp_widget;
						break;
					}
				}
				break;
			}
		}
		if (touch_window!=0xFFFFFFFF) {
			if (((touch_widget==NULL)||(!touch_widget->event_cb))&&(timer_check_event(TTOS_TIMER_DOUBLE_CLICK,false,false,0))) { // double click
				double_click=true;
			} else {
				last_widget=touch_widget;
				double_click=false;
				timer_check_event(TTOS_TIMER_DOUBLE_CLICK,true,true,500);
			}
		}

		return;
	}

	touch_counter=0;

	if (touch_window>=n_windows) {
		return;
	}

	if (sample.type==TTOS_TOUCH_NONE) {
		return;
	}

	if (sample.type==TTOS_TOUCH_CLICK) {
		if (double_click) {
			if (timer_check_event(TTOS_TIMER_DOUBLE_CLICK,false,false,0)) {
				timer_check_event(TTOS_TIMER_DOUBLE_CLICK,true,false,0);
				sample.type=TTOS_TOUCH_DOUBLE_CLICK;
			}
			timer_check_event(TTOS_TIMER_LONG_CLICK,true,false,0);
			double_click=false;
		}
		if (large_touch_made) {
			timer_check_event(TTOS_TIMER_LONG_CLICK,true,false,0);
			large_touch_made=false;
			sample.type=TTOS_TOUCH_NONE;
		} else {
			// remove any timer
			timer_check_event(TTOS_TIMER_LONG_CLICK,true,false,0);
			if ((touch_widget)&&(touch_widget->event_cb)) {
				if (main_adjust_coordinates(&sample,&sample2,touch_window,touch_widget)) {
					if (touch_widget->event_cb(&sample2,touch_widget)) {
						return;
					}
				}
			}
		}
	}

	// Send the unused event to the background manager

	if ((sample.type==TTOS_TOUCH_DRAG)||(sample.type==TTOS_TOUCH_END_DRAG)) {

		int32_t offset_x,offset_y;

		// remove any timer
		timer_check_event(TTOS_TIMER_LONG_CLICK,true,false,0);
		timer_check_event(TTOS_TIMER_DOUBLE_CLICK,true,false,0);
		double_click=false;

		if ((!windows[touch_window].fg_client)||(!windows[touch_window].fg_client->show_tab)||(!windows[touch_window].fg_client->show_tab->sensitive)) {
			return;
		}

		ctab=windows[touch_window].fg_client->show_tab;

		// This avoid a problem when a DRAG event starts in a tab, this one is destroyed, and continues in a new tab
		if (!ctab->valid_tmp) {
			ctab->xtmp=ctab->x_offset;
			ctab->ytmp=ctab->y_offset;
			ctab->valid_tmp=true;
		}

		offset_x=ctab->xtmp+sample.x-sample.ox;
		offset_y=ctab->ytmp+sample.y-sample.oy;

		if ((offset_x+ctab->minx)>0) {
			offset_x=-ctab->minx;
		}
		if ((offset_x+ctab->maxx)<windows[touch_window].w) {
			offset_x=windows[touch_window].w-ctab->maxx-1;
		}
		if ((offset_y+ctab->miny)>0) {
			offset_y=-ctab->miny;
		}
		if ((offset_y+ctab->maxy)<windows[touch_window].h) {
			offset_y=windows[touch_window].h-ctab->maxy-1;
		}
		ctab->x_offset=offset_x;
		ctab->y_offset=offset_y;

		if (sample.type==TTOS_TOUCH_DRAG) {
			struct pollfd st;

			st.events=POLLIN;
			st.fd=fd;
			st.revents=0;
			// if there are more events, they can be only TTOS_TOUCH_DRAG (omit, to avoid too many refreshes per second)
			// or TTOS_TOUCH_END_DRAG, which will be managed the next time the callback is called
			if (poll(&st,1,0)) {
				return;
			}
		}


		main_refresh(windows[touch_window].fg_client->conn);
		if (sample.type==TTOS_TOUCH_END_DRAG) {
			minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
			minibus_write_long(write_conn,TTOS_SR_TAB_DRAG);
			minibus_write_short(write_conn,sample.x-sample.ox); // send, first, the displacement made by the user
			minibus_write_short(write_conn,sample.y-sample.oy);
			minibus_write_short(write_conn,ctab->xtmp); // the original TAB's offset
			minibus_write_short(write_conn,ctab->ytmp);
			minibus_write_short(write_conn,ctab->x_offset); // and the final TAB's offset
			minibus_write_short(write_conn,ctab->y_offset);
			minibus_server_send(write_conn,ctab->client->conn,false);
		}

		return;
	}

	if ((sample.type==TTOS_TOUCH_DOUBLE_CLICK)&&(window_manager)) {
		minibus_start_message(write_conn,TTOS_SYSTEM_PORT_2,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_TOUCH_EVENT);
		minibus_write_byte(write_conn,TTOS_TS_MASTER);
		minibus_write_short(write_conn,0);
		minibus_write_short(write_conn,0);
		minibus_write_short(write_conn,sample.x);
		minibus_write_short(write_conn,sample.y);
		minibus_write_short(write_conn,sample.ox);
		minibus_write_short(write_conn,sample.oy);
		minibus_server_send(write_conn,window_manager->conn,false);
	}
}

uint32_t close_client(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *private_data) {

	struct minibus_data_element *data;
	int32_t error;
	uint32_t client_uid;
	struct minibus_list_element *element;
	struct _client_data *client;

	data=minibus_read_params(connection,"l",&error);

	if(!data) {
		return -1;
		switch(error) {
		case -1:
			other_send_error(connection,answer_port,TTOS_ERROR_NO_MEMORY);
			return -1;
		break;
		case -2:
			other_send_error(connection,answer_port,TTOS_ERROR_CANT_READ_DATA);
			return -1;
		break;
		case -3:
			other_send_error(connection,answer_port,TTOS_ERROR_INCORRECT_FORMAT);
			return -1;
		break;
		}
	}
	client_uid=data[0].value32;
	free(data);

	for(element=clients.list;element!=NULL;element=element->next) {
		client=element->data;
		if (client->uid==client_uid) {
			break;
		} else {
			client=NULL;
		}
	}

	if (client) {
		other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
		// Maximum priority
		minibus_start_message(write_conn,TTOS_SYSTEM_PORT_1,TTOS_NO_ANSWER);
		minibus_write_long(write_conn,TTOS_SR_EXIT);
		minibus_server_send(write_conn,client->conn,false);
	} else {
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_UID);
	}

	return 0;

}

void launch_command(char *command) {

	char *p1[2];

	if(0==fork()) {
		p1[0]=strdup(command);
		p1[1]=NULL;
		execve(command,p1,environ);
		exit(0);
	}


}

uint32_t default_callback(uint32_t port, uint32_t answer_port, struct minibus_connection *connection, void *data) {

	struct _client_data *client,*client2;
	struct minibus_list_element *element,*element2;
	struct ipc_bus_element *object;
	struct ipc_subscriber *subscriber;
	struct minibus_data_element msg_data;
	uint32_t retval;

	if ((port&0xFFFFF000)!=TTOS_IPC_UID_MASK) {
		other_send_error(connection,answer_port, TTOS_ERROR_INVALID_IPC_PORT);
		return 0; // no IPC port
	}

	client=minibus_get_connection_data(connection); // sender
	if (answer_port!=client->ipc_uid) {
		other_send_error(connection,answer_port, TTOS_ERROR_INVALID_IPC_ANSWER_PORT);
		return -1;
	}

	/* The message must start with a TTOS_SR_BUS_MESSAGE identifier, followed with a string */
	retval=minibus_read_data(connection,&msg_data);
	if ((retval)||(msg_data.type!=MINIBUS_TYPE_INT32)||(msg_data.value32!=TTOS_SR_BUS_MESSAGE)) {
		other_send_error(connection,answer_port, TTOS_ERROR_INVALID_IPC_MESSAGE_FORMAT);
		return -1;
	}

	retval=minibus_read_data(connection,&msg_data);
	if ((retval)||(msg_data.type!=MINIBUS_TYPE_STRING)) {
		other_send_error(connection,answer_port, TTOS_ERROR_INVALID_IPC_MESSAGE_FORMAT);
		return -1;
	}

	// check if this message is for a specific client
	for(element=clients.list;element;element=element->next) {
		client2=element->data;
		if (port==client2->ipc_uid) {
			minibus_server_send(connection,client2->conn,true);
			other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
			return 0;
		}
	}

	object=NULL;
	for(element=ipc.list;element;element=element->next) {
		object=element->data;
		if (port==object->port) {
			break;
		}
		object=NULL;
	}

	if (!object) { // port unassigned; error
		other_send_error(connection,answer_port,TTOS_ERROR_INVALID_BUS);
		return -1;
	}

	if (object->owner!=client->ipc_uid) {

		/* If bus is ALL_TO_OWNER and the sender is not the owner, send the message to the owner */
		if (object->flags&TTOS_IPC_ALL_TO_OWNER) {
			minibus_server_send(connection,object->connection,true);
			other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
			return 0;
		}

		/* If not, the user is not allowed to send messages to this port */
		other_send_error(connection,answer_port,TTOS_ERROR_NOT_ALLOWED_TO_SEND);
		return -1;
	}

	/* The owner is not allowed to send a message to a bus which will redirect all messages to itself */
	if (object->flags&TTOS_IPC_ALL_TO_OWNER) {
		other_send_error(connection,answer_port,TTOS_ERROR_NOT_ALLOWED_TO_SEND);
		return -1;
	}

	/* If the code reaches here, then this is an OWNER_TO_ALL bus, and the onwer is sending a message */
	for(element2=object->subscribers.list;element2;element2=element2->next) {
		subscriber=element2->data;
		minibus_server_send(connection,subscriber->conn,true);
	}
	other_send_error(connection,answer_port,TTOS_ERROR_NO_ERROR);
	return 0;
}

void print_help() {

	printf("\nUsage: ttos [-big] [-normal] [-cw] [-ccw] [-udown] [-portrait] [-landscape]");

}

int main(int argc,char **argv) {

	struct minibus_list_element *list_element;
	int param;
	char big;
	char rotation;
	char portrait;

	ttos_uids=uid_init();
	if (!ttos_uids) {
		printf("Can't initializate the UID generator. Aborting\n");
	}
	ttos_ipc_uid=uid_init();
	if (!ttos_ipc_uid) {
		printf("Can't initializate the UID generator. Aborting\n");
	}

	param=1;
	big=0;
	portrait=PORTRAIT_MODE;
	rotation=SCREEN_ROTATE;

	while(param<argc) {
		if(!strcmp(argv[param],"-big")) {
			big=1;
			param++;
			continue;
		}
		if(!strcmp(argv[param],"-normal")) {
			rotation=0;
			param++;
			continue;
		}
		if(!strcmp(argv[param],"-cw")) {
			rotation=1;
			param++;
			continue;
		}
		if(!strcmp(argv[param],"-udown")) {
			rotation=2;
			param++;
			continue;
		}
		if(!strcmp(argv[param],"-ccw")) {
			rotation=3;
			param++;
			continue;
		}
		if(!strcmp(argv[param],"-portrait")) {
			portrait=1;
			param++;
			continue;
		}
		if(!strcmp(argv[param],"-landscape")) {
			portrait=0;
			param++;
			continue;
		}
		print_help();
		exit(0);
	}

	if (big) {
		base=ttos_init_screen(rotation,portrait,480,272);
	} else {
		base=ttos_init_screen(rotation,portrait,320,240);
	}
	ts=ttos_init_touchscreen();
	touch_counter=0;
	large_touch_made=false;

	windows=(struct _window*)malloc(sizeof(struct _window));
	if (!windows) {
		printf("Can't generate the window list. Aborting\n");
		exit(1);
	}

	windows[0].ox=0;
	windows[0].oy=0;
	windows[0].mx=base->w-1;
	windows[0].my=base->h-1;
	windows[0].w=base->w;
	windows[0].h=base->h;
	windows[0].fg_client=NULL;
	windows[0].visible=true;
	n_windows=1;

	minibus_list_init(&clients,clients_destroy,NULL);
	minibus_list_init(&timers,timer_destroy,timer_compare);
	minibus_list_init(&tab_scroll,tabs_free_tab_scroll,NULL);
	minibus_list_init(&bitmaps,bitmaps_free_container,NULL);
	minibus_list_init(&ipc,ipc_free_bus,NULL);
	window_manager=NULL;

	base_font=fonts_init((int8_t *)"./freesans.font");

	main_queue=minibus_init("/tmp/ttos_queue",true);
	if (main_queue==NULL) {
		printf("Can't connect to the queue. Aborting\n");
		exit(1);
	}

	write_conn=main_queue->send_connection;

	minibus_set_new_connection_callback(main_queue,cb_new_connection,NULL);
	minibus_set_closed_connection_callback(main_queue,cb_closed_connection,NULL);
	minibus_add_callback(main_queue,TTOS_SM_INIT,clients_init,base);
	minibus_add_callback(main_queue,TTOS_SM_DEFINE_WINDOWS,clients_define_windows,NULL);
	minibus_add_callback(main_queue,TTOS_SM_WINDOW_SET_VISIBLE,clients_window_set_visible,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_CLIENT_WINDOW,clients_set_client_window,NULL);
	minibus_add_callback(main_queue,TTOS_SM_MOVE_CLIENT_TO_FG,clients_move_to_fg,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_TAB,tabs_create_tab,NULL);
	minibus_add_callback(main_queue,TTOS_SM_GET_CURRENT_TAB,tabs_get_current_tab,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SELECT_TAB,tabs_select_show_tab,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_TAB_BG_COLOR,tabs_set_bg_color,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_TAB_OFFSET,tabs_set_tab_offset,NULL);
	minibus_add_callback(main_queue,TTOS_SM_DESTROY_TAB,tabs_destroy_tab,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_TAB_SIZE,tabs_set_size,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_TAB_MARGINS,tabs_set_margins,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_WIDGET_VISIBILITY,widgets_set_visibility,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_WIDGET_POSITION,widgets_set_position,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_WIDGET_PICTURE,widgets_create_picture,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_PICTURE,widgets_set_picture,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_WIDGET_LABEL,widgets_create_label,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SET_LABEL_TEXT,widgets_set_label_text,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_WIDGET_BUTTON,widgets_create_button,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_WIDGET_COLOR_AREA,widgets_create_color_area,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_WIDGET_DRAWABLE,widgets_create_drawable,NULL);
	minibus_add_callback(main_queue,TTOS_SM_UPDATE_DRAWABLE,widgets_update_drawable,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_SHAREMEM,sharemem_create,NULL);
	minibus_add_callback(main_queue,TTOS_SM_DESTROY_SHAREMEM,sharemem_destroy,NULL);
	minibus_add_callback(main_queue,TTOS_SM_REPAINT_SCREEN,main_refresh_cb,NULL);
	minibus_add_callback(main_queue,TTOS_SM_NEW_TIMER,timer_new_timer,NULL);
	minibus_add_callback(main_queue,TTOS_SM_DELETE_TIMER,timer_delete_timer,NULL);
	minibus_add_callback(main_queue,TTOS_SM_CLOSE_CLIENT,close_client,NULL);
	minibus_add_callback(main_queue,TTOS_SM_REGISTER_BUS,ipc_register_bus,NULL);
	minibus_add_callback(main_queue,TTOS_SM_UNREGISTER_BUS,ipc_unregister_bus,NULL);
	minibus_add_callback(main_queue,TTOS_SM_SUBSCRIBE_BUS,ipc_subscribe_to_bus,NULL);
	minibus_add_callback(main_queue,TTOS_SM_UNSUBSCRIBE_BUS,ipc_unsubscribe_to_bus,NULL);
	minibus_add_listener(main_queue,ts_fd(ts),POLLIN,touchscreen_cb,NULL);
	minibus_set_default_callback(main_queue,default_callback,NULL);
	minibus_set_timeout_callback(main_queue,timer_timeout_callback,NULL);

	launch_command("ttos_manager.bin");

	minibus_run_loop(main_queue);

	/* debug code to ensure that there are no memory leaks (to be used with valgrind and FAKE_FB, of course) */

	minibus_remove_callback(main_queue,TTOS_SM_INIT);
	minibus_remove_callback(main_queue,TTOS_SM_DEFINE_WINDOWS);
	minibus_remove_callback(main_queue,TTOS_SM_WINDOW_SET_VISIBLE);
	minibus_remove_callback(main_queue,TTOS_SM_SET_CLIENT_WINDOW);
	minibus_remove_callback(main_queue,TTOS_SM_MOVE_CLIENT_TO_FG);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_TAB);
	minibus_remove_callback(main_queue,TTOS_SM_GET_CURRENT_TAB);
	minibus_remove_callback(main_queue,TTOS_SM_SELECT_TAB);
	minibus_remove_callback(main_queue,TTOS_SM_SET_TAB_BG_COLOR);
	minibus_remove_callback(main_queue,TTOS_SM_SET_TAB_OFFSET);
	minibus_remove_callback(main_queue,TTOS_SM_DESTROY_TAB);
	minibus_remove_callback(main_queue,TTOS_SM_SET_TAB_SIZE);
	minibus_remove_callback(main_queue,TTOS_SM_SET_TAB_MARGINS);
	minibus_remove_callback(main_queue,TTOS_SM_SET_WIDGET_VISIBILITY);
	minibus_remove_callback(main_queue,TTOS_SM_SET_WIDGET_POSITION);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_WIDGET_PICTURE);
	minibus_remove_callback(main_queue,TTOS_SM_SET_PICTURE);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_WIDGET_LABEL);
	minibus_remove_callback(main_queue,TTOS_SM_SET_LABEL_TEXT);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_WIDGET_BUTTON);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_WIDGET_COLOR_AREA);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_WIDGET_DRAWABLE);
	minibus_remove_callback(main_queue,TTOS_SM_UPDATE_DRAWABLE);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_SHAREMEM);
	minibus_remove_callback(main_queue,TTOS_SM_DESTROY_SHAREMEM);
	minibus_remove_callback(main_queue,TTOS_SM_REPAINT_SCREEN);
	minibus_remove_callback(main_queue,TTOS_SM_NEW_TIMER);
	minibus_remove_callback(main_queue,TTOS_SM_DELETE_TIMER);
	minibus_remove_callback(main_queue,TTOS_SM_CLOSE_CLIENT);
	minibus_remove_callback(main_queue,TTOS_SM_REGISTER_BUS);
	minibus_remove_callback(main_queue,TTOS_SM_UNREGISTER_BUS);
	minibus_remove_callback(main_queue,TTOS_SM_SUBSCRIBE_BUS);
	minibus_remove_callback(main_queue,TTOS_SM_UNSUBSCRIBE_BUS);
	//minibus_remove_callback(main_queue,);

	for(list_element=main_queue->conections.list;list_element!=NULL;list_element=list_element->next) {
		printf("Conections\n");
	}

	for(list_element=main_queue->callbacks.list;list_element!=NULL;list_element=list_element->next) {
		printf("Callbacks\n");
	}

	minibus_list_free(&timers);
	minibus_list_free(&ipc);
	minibus_release(main_queue);
	free(windows);
	fonts_destroy(base_font);
	ttos_destroy_touchscreen(ts);
	ttos_destroy_screen(base);
	uid_destroy(ttos_uids);
	uid_destroy(ttos_ipc_uid);
	return 0;
}
