/*
 * Copyright 2011 Raster Software Vigo
 *
 * This file is part of ttos
 *
 * ttos is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * ttos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * other.c
 *
 *  Created on: 18/09/2010
 *      Author: raster
 */

#include <stdlib.h>

#include "minibus/minibus.h"
#include "global_vars.h"
#include "other.h"


/**
 * Returns a pointer to a client's structure given its UID
 *
 * @param uid The client's UID
 *
 * @return The client's structure, or NULL if there's no client with that UID
 */

struct _client_data *other_get_client_by_uid(uint32_t uid) {

	struct _client_data *client_element;
	struct minibus_list_element *element;

	for(element=clients.list;element!=NULL;element=element->next) {
		client_element=element->data;
		if (uid==minibus_get_uid(client_element->conn)) {
			return (client_element);
		}
	}
	return (NULL);
}


/**
 * Checks if a connection belongs to the window manager or not
 *
 * @param connection The connection
 *
 * @return True if the connection belongs to the window manager client; False if not.
 */

bool other_client_is_wm(struct minibus_connection *connection) {

	struct _client_data *client_element;

	client_element=minibus_get_connection_data(connection);
	return (client_element->window_manager);

}

/**
 * Sends an error message to an specified port. It won't have an answer port, and will have two longs as data: the first
 * one, the TTOS_SR_ERROR identifier, and the second one, the error code itself
 *
 * @param connection The client's connection
 * @param port The client's port to which send the error message
 * @param error The error code
 */

void other_send_error(struct minibus_connection *connection, uint32_t port, uint32_t error) {

	if (port==TTOS_NO_ANSWER) {
		return; // port 0xFFFFFFFF means "don't send errors"
	}

	minibus_start_message(write_conn,port,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_ERROR);
	minibus_write_long(write_conn,error);
	minibus_server_send(write_conn,connection,false);

}


/**
 * Sends an UID message to an specified port. It won't have an answer port, and will have two longs as data: the first
 * one, the TTOS_SR_NEW_WIDGET identifier, and the second one, the UID itself
 *
 * @param connection The client's connection
 * @param port The client's port to which send the error message
 * @param uid The element UID
 */

void other_send_uid(struct minibus_connection *connection, uint32_t port, uint32_t uid) {

	if (port==TTOS_NO_ANSWER) {
		return; // port 0xFFFFFFFF means "don't send errors"
	}

	minibus_start_message(write_conn,port,TTOS_NO_ANSWER);
	minibus_write_long(write_conn,TTOS_SR_NEW_WIDGET);
	minibus_write_long(write_conn,uid);
	minibus_server_send(write_conn,connection,false);

}


/**
 * Applies to a given color the specified alpha value, and returns it
 *
 * @param color A pointer to the color to which apply the alpha value. It must be NOT premultiplied
 * @param rcolor A pointer to a structure where to return the new color
 * @param alpha The alpha value
 *
 * @return false on success, true on error
 */

bool other_apply_alpha(struct ttos_color *color, struct ttos_color *rcolor, uint8_t alpha) {

	uint16_t a;
	if ((!color)||(!rcolor)) {
		return true;
	}

	a=((((uint16_t)alpha)*((uint16_t)color->a))>>8);
	rcolor->a=(uint8_t)a;
	rcolor->r=(uint8_t)((a*((uint16_t)color->r))>>8);
	rcolor->g=(uint8_t)((a*((uint16_t)color->g))>>8);
	rcolor->b=(uint8_t)((a*((uint16_t)color->b))>>8);
	return (false);
}
