/*
 * Copyright 2010 (C) Raster Software Vigo (Sergio Costas)
 * This file is part of TTOS
 *
 * TTOS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * TTOS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * portable.h
 *
 *  Created on: 01/09/2010
 *      Author: raster
 */

#include <stdint.h>
#include <stdbool.h>
#include <sys/time.h>
#include "blit.h"

#ifndef SCREEN_H_
#define SCREEN_H_

/* Variables for screen.
 *
 * SCREEN_WIDTH and SCREEN_HEIGHT is the resolution, as seen by the programs
 * SCREEN_ROTATE is the desired rotation:
 * 0: normal
 * 1: clockwise
 * 2: upside down
 * 3: counterclockwise
 *
 * SCREEN_COLOR is the color depth in bytes
 * SCREEN_TOTAL is the screen size in bytes (WIDTH*HEIGHT*COLOR)
 */

#ifdef TTOS_USE_FB
#define SCREEN_ROTATE 0
#define PORTRAIT_MODE 0
#endif

#ifdef TTOS_USE_TSLIB
#include "include/tslib.h"
#endif

#if defined(TTOS_USE_SDL) || defined(TTOS_USE_FAKE_FB)
#define SCREEN_WIDTH 320
#define SCREEN_HEIGHT 240
#define SCREEN_ROTATE 0
#define PORTRAIT_MODE 0

struct tsdev {
	int fd;
};

int ts_fd(struct tsdev *);
#endif

enum TTOS_TOUCH_TYPES {TTOS_TOUCH_START_CLICK, TTOS_TOUCH_CLICK, TTOS_TOUCH_DOUBLE_CLICK, TTOS_TOUCH_LARGE_CLICK, TTOS_TOUCH_DRAG, TTOS_TOUCH_END_DRAG, TTOS_TOUCH_NONE };

#define MAX_DANDD 5

struct ttos_tssample {
	int		x;
	int		y;
	int     ox;
	int     oy;
	struct timeval tv;
	enum TTOS_TOUCH_TYPES type;
};

TTOS_Surface *ttos_init_screen(uint8_t rotation, uint8_t portrait, uint16_t width,uint16_t height);
void ttos_destroy_screen(TTOS_Surface *surface);
struct tsdev *ttos_init_touchscreen();
void ttos_destroy_touchscreen(struct tsdev *object);
uint32_t ttos_dump_screen(TTOS_Surface *surface);
struct tsdev *ttos_init_touchscreen();
int32_t ttos_read_ts(struct tsdev *ts, struct ttos_tssample *sample);

#endif /* SCREEN_H_ */
